#! /usr/bin/env bats

# This is a test file for use with the bats testing framework
# https://github.com/bats-core/bats-core
# On Debian: apt install bats

function setup {
    ORIGDIR="$(pwd)"

    # If no $FORGEBUILD is supplied in ENV, use the default
    # (otherwise you can't `bats ./tests/` manually)
    [ -z "$FORGEBUILD" ] && FORGEBUILD="forgebuild"

    # Ensure output from previous test does not exist
    # (in case something crashed)
    clean

    # Setup simple git repository
    GITREPO="$(mktemp -d)"
    cd $GITREPO
    git init
    echo "first commit" > README
    git add README
    git commit -m "first commit"
    
    # Setup forgebuild tasks
    FORGEBUILDDIR="$(mktemp -d)"
    echo "$GITREPO" > $FORGEBUILDDIR/time.source
    echo "date +%s%N > /tmp/forgebuild-time" > $FORGEBUILDDIR/time
    chmod +x $FORGEBUILDDIR/time
    echo "$GITREPO" > $FORGEBUILDDIR/host.source
    echo "cat \$GITBUILDCONF/entry > /tmp/forgebuild-host" > $FORGEBUILDDIR/host
    chmod +x $FORGEBUILDDIR/host
    echo "$GITREPO" > $FORGEBUILDDIR/branch.source
    echo "cat \$GITBUILDDIR/.branch/branch > /tmp/forgebuild-branch" > $FORGEBUILDDIR/branch
    chmod +x $FORGEBUILDDIR/branch
    
    # Setup settings
    mkdir $FORGEBUILDDIR/config
    echo "default" > $FORGEBUILDDIR/config/entry
    mkdir $FORGEBUILDDIR/$HOSTNAME
    echo "host" > $FORGEBUILDDIR/$HOSTNAME/entry
}

function teardown {
    cd $ORIGDIR
    clean
}

function clean {
    if [ -d $GITREPO ]; then rm -rf $GITREPO; fi
    if [ -d $FORGEBUILDDIR ]; then rm -rf $FORGEBUILDDIR; fi
    if [ -f /tmp/forgebuild-time ]; then rm /tmp/forgebuild-time; fi
    if [ -f /tmp/forgebuild-host ]; then rm /tmp/forgebuild-host; fi
    if [ -f /tmp/forgebuild-branch ]; then rm /tmp/forgebuild-branch; fi
}

# Simple repository

@test "repository is cloned" {
    $FORGEBUILD -b $FORGEBUILDDIR time
    [ -f $FORGEBUILDDIR/.time/README ]
}

@test "first run triggers task" {
    $FORGEBUILD -b $FORGEBUILDDIR time
    [ -f /tmp/forgebuild-time ]
}

@test "no update does not trigger task" {
    $FORGEBUILD -b $FORGEBUILDDIR time
    [ -f /tmp/forgebuild-time ]
    TIME="$(cat /tmp/forgebuild-time)"
    $FORGEBUILD -b $FORGEBUILDDIR time
    NEWTIME="$(cat /tmp/forgebuild-time)"
    [[ "$TIME" = "$NEWTIME" ]]
}

@test "update triggers task" {
    $FORGEBUILD -b $FORGEBUILDDIR time
    [ -f /tmp/forgebuild-time ]
    TIME="$(cat /tmp/forgebuild-time)"
    cd $GITREPO
    touch NEWFILE
    git add NEWFILE
    git commit -m "second commit"

    $FORGEBUILD -b $FORGEBUILDDIR time
    NEWTIME="$(cat /tmp/forgebuild-time)"

    [[ "$TIME" != "$NEWTIME" ]]
}

# Task settings

@test "host-based settings are loaded" {
    $FORGEBUILD -b $FORGEBUILDDIR host
    [[ "$(cat /tmp/forgebuild-host)" = "host" ]]
}

@test "unknown host uses default settings" {
    HOST="UNKNOWN" $FORGEBUILD -b $FORGEBUILDDIR host
    [[ "$(cat /tmp/forgebuild-host)" = "default" ]]
}

@test "can checkout to a specific branch/commit" {
    echo "branch" > $FORGEBUILDDIR/branch.checkout
    cd $GITREPO
    git checkout -b branch
    echo "branch" > branch
    git add branch
    git commit -m "new branch"
    # TODO: should be able to checkout to the default branch even when it's not called master
    git checkout master
    $FORGEBUILD -b $FORGEBUILDDIR branch
    [[ "$(cat /tmp/forgebuild-branch)" = "branch" ]]
}

@test "task only runs in allowlisted hosts" {
    # When t.hosts exist, only this (line-separated) list of hosts
    # runs the task
    echo "$HOSTNAME" > $FORGEBUILDDIR/time.hosts
    $FORGEBUILD -b $FORGEBUILDDIR time
    TIME="$(cat /tmp/forgebuild-time)"

    # Doesn't run on unknown host
    HOST="UNKNOWN" $FORGEBUILD -b $FORGEBUILDDIR time
    NEWTIME="$(cat /tmp/forgebuild-time)"
    [[ "$TIME" = "$NEWTIME" ]]

    echo "TESTHOST" >> $FORGEBUILDDIR/time.hosts
    # Force trigger because repo was not updated
    HOST="TESTHOST" $FORGEBUILD -f -b $FORGEBUILDDIR time
    NEWTIME="$(cat /tmp/forgebuild-time)"
    [[ "$TIME" != "$NEWTIME" ]]
}

@test "task doesn't run in ignored host" {
    touch $FORGEBUILDDIR/$HOSTNAME/time.ignore
    $FORGEBUILD -b $FORGEBUILDDIR time
    [ ! -f /tmp/forgebuild-time ]

    HOST="TESTHOST" $FORGEBUILD -f -b $FORGEBUILDDIR time
    [ -f /tmp/forgebuild-time ]
}

@test "repository unavailable fails to trigger task" {
    rm -Rf $GITREPO
    ! $FORGEBUILD -b $FORGEBUILDDIR time
    [ ! -f /tmp/forgebuild-time ]
}

# Repository with submodules

function setup_submodule {
    SUBMODULE="$(mktemp -d)"
    cd $SUBMODULE
    git init
    echo "SUB" > README
    git add README
    git commit -m "first commit"

    cd $GITREPO
    git submodule add $SUBMODULE sub
    git commit -m "add submodule"
}

@test "submodule is cloned" {
    setup_submodule
    $FORGEBUILD -b $FORGEBUILDDIR time
    [ -f $FORGEBUILDDIR/.time/sub/README ]
    [[ "$(cat $FORGEBUILDDIR/.time/sub/README)" = "SUB" ]]
}

@test "first submodule run triggers task" {
    setup_submodule
    $FORGEBUILD  -b $FORGEBUILDDIR time
    [ -f /tmp/forgebuild-time ]
}

@test "submodule update does not trigger task by default" {
    setup_submodule
    $FORGEBUILD -b $FORGEBUILDDIR time
    [ -f /tmp/forgebuild-time ]
    TIME="$(cat /tmp/forgebuild-time)"

    cd $SUBMODULE
    echo "new stuff" > NEWFILE
    git add NEWFILE
    git commit -m "second commit"

    $FORGEBUILD -b $FORGEBUILDDIR time
    NEWTIME="$(cat /tmp/forgebuild-time)"

    [[ "$TIME" = "$NEWTIME" ]]
}

@test "submodule update triggers task when task.subupdates is set" {
    setup_submodule
    touch $FORGEBUILDDIR/time.subupdates
    $FORGEBUILD -b $FORGEBUILDDIR time
    [ -f /tmp/forgebuild-time ]
    TIME="$(cat /tmp/forgebuild-time)"

    cd $SUBMODULE
    echo "new stuff" > NEWFILE
    git add NEWFILE
    git commit -m "second commit"

    $FORGEBUILD -b $FORGEBUILDDIR time
    NEWTIME="$(cat /tmp/forgebuild-time)"

    [[ "$TIME" != "$NEWTIME" ]]
}

# CLI arguments/flags

@test "running specific tasks only runs those tasks" {
    $FORGEBUILD -b $FORGEBUILDDIR time host
    [ -f /tmp/forgebuild-time ]
    [ -f /tmp/forgebuild-host ]
    [ ! -f /tmp/forgebuild-branch ]
}

@test "not specifying tasks runs all tasks" {
    $FORGEBUILD -b $FORGEBUILDDIR
    [ -f /tmp/forgebuild-time ]
    [ -f /tmp/forgebuild-host ]
    [ -f /tmp/forgebuild-branch ]
}

@test "forced run triggers task" {
    $FORGEBUILD -b $FORGEBUILDDIR time
    [ -f /tmp/forgebuild-time ]
    TIME="$(cat /tmp/forgebuild-time)"
    $FORGEBUILD -f -b $FORGEBUILDDIR time
    NEWTIME="$(cat /tmp/forgebuild-time)"
    [[ "$TIME" != "$NEWTIME" ]]
}

@test "can run with relative basedir" {
    cd $FORGEBUILDDIR
    $FORGEBUILD -b . time
    [ -f /tmp/forgebuild-time ]
}

@test "can run with trailing slash in basedir" {
    cd $FORGEBUILDDIR
    $FORGEBUILD -b ./ time
    [ -f /tmp/forgebuild-time ]
}

@test "can run with ~/.forgebuild" {
    if [ -d $HOME/.forgebuild ]; then
        NEWDIR="$(mktemp -d)"
        mv $HOME/.forgebuild $NEWDIR;
    fi
    ln -s $FORGEBUILDDIR $HOME/.forgebuild
    $FORGEBUILD time
    [ -f /tmp/forgebuild-time ]
    rm $HOME/.forgebuild
    if [ ! -z $NEWDIR ]; then
        mv $NEWDIR/.forgebuild $HOME
    fi
    rmdir $NEWDIR
}
